import re
import os
import datetime
import tempfile
import subprocess
from .constants import SPECIAL_DATES_DISPLAY


def filter_to_string(filters, search_query):
    pieces = []
    project = filters.get("project")
    status = filters.get("status")
    if project:
        pieces += [f" project:{project}"]
    if status:
        pieces += [f" status:{status}"]
    if search_query:
        pieces += [f" search:'{search_query}"]
    return "".join(pieces)


def remove_rich_tag(text):
    """remove rich styling from a string"""
    pattern = r"\[[^\]]*\](.*?)\[/\]"
    return re.sub(pattern, r"\1", text)


def advance_enum_val(enum_type, cur_val):
    """advance a value in an enum by one, wrapping around"""
    members = [str(e.value) for e in enum_type]
    cur_idx = members.index(remove_rich_tag(cur_val))
    next_idx = (cur_idx + 1) % len(members)
    return members[next_idx]


def get_color_enum(value: str, enum: dict[str, dict]) -> str:
    color = enum.get(value, {"color": "#ff0000"})["color"]
    return f"[{color}]{value}[/]"


def get_colored_date(date: datetime.date) -> str:
    if isinstance(date, datetime.date):
        as_date = date
        as_str = date.strftime("%Y-%m-%d")
    elif isinstance(date, str):
        as_date = datetime.datetime.strptime(date, "%Y-%m-%d").date()
        as_str = as_date.strftime("%Y-%m-%d")
    else:
        return "~bad~"

    if as_str in SPECIAL_DATES_DISPLAY:
        return SPECIAL_DATES_DISPLAY[as_str]
    today = datetime.date.today()
    if as_date < today:
        return f"[#eeeeee on #dd1111]{as_str}[/]"
    else:
        # Calculate weeks into future
        delta = as_date - today
        weeks = delta.days // 7

        colors = [
            # "#FF4000",
            "#FF8000",
            "#FFA533",
            "#FFD24D",
            "#FFE680",
            "#FFF4B3",
            "#FFFFFF",
        ]

        if weeks >= len(colors):
            color_index = -1
        else:
            color_index = weeks
        color = colors[color_index]
    return f"[{color}]{as_str}[/]"


def get_text_from_editor(initial_text: str = "") -> str | None:
    editor = os.environ.get("EDITOR", "vim")

    with tempfile.NamedTemporaryFile(suffix=".txt", mode="w+", delete=True) as tf:
        tf.write(initial_text)
        tf.flush()

        initial_mtime = os.path.getmtime(tf.name)
        subprocess.call([editor, tf.name])

        if os.path.getmtime(tf.name) == initial_mtime:
            return None

        with open(tf.name, "r") as f:  # Open fresh to get new content
            edited_text = f.read()
            return edited_text
